// Copyright 2022 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
// Modified by Jeremy Retailleau.

#include "./defineResolver.h"

namespace pxr {

Ar_ResolverFactoryBase::~Ar_ResolverFactoryBase() = default;

}  // namespace pxr
