// Copyright 2020 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
// Modified by Jeremy Retailleau.

#include "./writableAsset.h"

namespace pxr {

ArWritableAsset::ArWritableAsset() = default;

ArWritableAsset::~ArWritableAsset() = default;

}  // namespace pxr
